#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#define NAME "getOptOSPFWeights"
#include "topo_help.h"

/* 
 * Usage: ./getOptOSPFWeights [-m|-v] topo tmdir fileTag numTMs [outdir]
 *
 *  Given a topology and a certain number of TMs finds optimal weights.
 *
 * topo is a generic topo file
 * tmdir is location of tms
 * optimize over TMs in tmdir/fileTag_%d    for 1, ..., numTMs
 * output goes into outdir if present, else currentDirector
 * outputfile will be tagged 
 *             fileTag_%d.weights for indiv
 *             fileTag_%d_%d.weights for multiple TMs at the same time
 *             fileTag_%d_%d.fail_weights for indiv, also optimize over failures
 * use -m to yield a single weight-set over all TMs
 * use -v to simultaneously optimize over failures
 *
 */

#define OUTDIRSIZE 300
int main(int argc, char **argv){
  FILE* fin, *fout;
  int numnodes,linknum;
  char **topomatrix, *type_list, getoptch;
  Link*** linkmatrix;
  int switch_opt_one_at_a_time=1, switch_optimize_over_fails=0;
  
  if(argc < 5){
    printf("Usage: %s [-m|-v] <topo> <tmdir> <fileTag> <numTMs> [OUTDIR]\n", NAME);
    exit(0);
  }
  
  while( (getoptch = getopt(argc, argv, "mv")) != EOF) {
    if(DEBUG) printf("found an option, %c, optind=%d\n", getoptch, optind);
    switch( getoptch ) {
    case 'm':
      switch_opt_one_at_a_time = 0;
      break;
    case 'v':
      switch_optimize_over_fails = 1;
      break;
    default:
      printf ("Unknown option, ignore\n");
    }
  }

  if (DEBUG)
    printf("outside optind is %d, argc is %d, opt is %s\n", optind, argc, argv[optind]);

  if ( argc-optind < 4 || argc-optind > 5){
    printf("Usage: %s [-i|-v] <topo> <tmdir> <fileTag> <numTMs> [OUTDIR]\n", NAME);
    exit(0);
  }

  char *topoFile = argv[optind], *tmDir = argv[optind+1], *fileTag = argv[optind+2];
  int how_many_tms = atoi(argv[optind+3]);
  if ( how_many_tms < 1 ) {
    printf("invalid # of tms %d\n", how_many_tms); exit(437);
  }


  char outdir [OUTDIRSIZE];
  if ( argc - optind == 5 ){
    if ( strlen(argv[optind+4]) > OUTDIRSIZE - 1 )
      { printf("too big outdirname\n"); exit(436); }
    strcpy(outdir, argv[optind+4]);
  }else
    strcpy(outdir, ".");
  
  printf("getOptOSPFWeights: \n\tTopo %s\n\tTMs from Dir %s\n\tFileTag %s\n\tTMs 1-%d\n\tOutputDir %s\n\tFlags: One-by-one %s, Optimize-Fails %s\n",
	 topoFile, tmDir, fileTag, how_many_tms, outdir, 
	 switch_opt_one_at_a_time?"Yes":"No",
	 switch_optimize_over_fails?"Yes":"No" );

  if(switch_optimize_over_fails){
    if (how_many_tms > 1 && !switch_opt_one_at_a_time)
      printf("\n\tOptimizing over failures not built for multiple TMS; Ignore optimize_over_fails\n");
    else    printf("\n\tOptimizing over 1 failures\n");
  }
  if(how_many_tms > 1 && switch_opt_one_at_a_time)
    printf("\n\tOptimizing over singleton TMs\n");
  else printf("\n\tOptimizing over multiple TMs\n");
    


  // open topo
  fin = fopen(topoFile, "r");
  if( fin == NULL ) { printf("Cannot open %s\n", topoFile); exit(1); }
  numnodes = readtopo(fin, &topomatrix, &linkmatrix, &type_list, &linknum);

  // outFilePrefix
  char outFilePrefix[OUTDIRSIZE+100];
  sprintf(outFilePrefix, "%s/%s", outdir, fileTag);

  {
    double **dist;
    char ***succ;
    int ind1; 
    if(switch_opt_one_at_a_time) {
      double **tm = get2ddouble(numnodes);
      for(ind1=0; ind1<how_many_tms; ind1++){
	readTMs(numnodes, tmDir, ind1+BASETM, tm, fileTag);
	setoptimalweights(&tm, numnodes, linknum, linkmatrix, 
			  topomatrix, ind1+BASETM, 1, switch_optimize_over_fails, outFilePrefix);
      }
    }else {
      double ***tms = (double ***) malloc(how_many_tms * sizeof(double**));
      for(ind1=0; ind1<how_many_tms; ind1++){
	tms[ind1] = get2ddouble(numnodes);
	readTMs(numnodes, tmDir, BASETM+ind1, tms[ind1], fileTag);
      }
      setoptimalweights(tms, numnodes, linknum, linkmatrix, topomatrix, 
			BASETM, how_many_tms, 0, outFilePrefix);
    }
  }
}
