#define MAXLINELEN 1500

#ifndef DEBUG
#define DEBUG 0
#endif

#define MAXNUMNODES 100
#define MAXLINKNUM  200
#define MAXPATHSPERPAIR 100
#define PREDSTRING_LEN 100

#define CC_CAP 10000 /* Mbps */
#define CP_CAP 2500  /* Mbps */
#define PP_CAP 2500  /* Mbps */

#define NUMTMS 1
#define BASETM 1
#define NUMFAILURES 1
#define NUMREPEATS 2

#define OPTIMIZE_OVER_FAILURES 0
#define UPDATE_CRITICAL_SET_FREQ 10
#define CRITICAL_SET_SIZE 5
extern char* optarg;
extern int optind;

#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

typedef struct link{
  int id;
  double lat;
  int cap; /* Mbps */
} Link;

/* node in a linked list of paths */
typedef struct path{
  int path_vector[MAXNUMNODES]; 
  int numhops;
  double lat;
  struct path* next; 
  struct path* snext;
} Path;

typedef struct node{
  int id;
  double x;
  double y;
} Node;

inline int is_in( int a, int min, int max) {
  return ((a>=min) && (a<= max));
}

void free2ddouble(int n, double** ptr){
  int ind1;
  for(ind1=0; ind1<n; ind1++)
    free(ptr[ind1]);
  free(ptr);
  return;
}

void **copy2ddouble(int n, double**dest, double**src){
  int ind1;
  for(ind1=0; ind1<n; ind1++)
    memcpy(dest[ind1], src[ind1], n*sizeof(double));
}

void print2ddouble(int n, double **mat){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++)
      printf("%g ", mat[ind1][ind2]);
    printf("\n");
  }
}

double ** get2ddouble(int n){
  int ind1;
  double ** dummy = (double **) calloc(n, sizeof(double*));
  for(ind1=0; ind1<n; ind1++){
    dummy[ind1] = (double*) calloc(n, sizeof(double));
    bzero(dummy[ind1], n * sizeof(double));
  }
  return dummy;
}

void add2ddouble(int n, double **dest, double **src){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      dest[ind1][ind2] += src[ind1][ind2];
  return;
}
void divide2ddouble(int n, double **dest, double d){
  if( d <= 0) { printf("cannot divide by zero! "); exit(22); }
  int ind1, ind2;

  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      dest[ind1][ind2] /= d;
  return;
}

void get_bimodal_tm(int n, Link*** linkmatrix, int l, 
		    FILE* f, double**demandmatrix) {
  double weights[MAXNUMNODES][MAXNUMNODES];
  int ind1, ind2;

  if(f) fprintf(f, "Demands = [\n");
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++){
      if(ind1 == ind2) 
	weights[ind1][ind2] = 0;
      else 
	{
	  double x1 = drand48(), x2=drand48();
	  if(x1 > .5) {
	    if (DEBUG) printf("-");
	    weights[ind1][ind2] = x2*.1 + .8;
	  }
	  else{
	    if ( DEBUG) printf("+");
	    weights[ind1][ind2] = x2*.1 + .2;
	  }
	}
      if(f) fprintf(f, " %-9.4f", weights[ind1][ind2]*100);
      if(demandmatrix) (demandmatrix)[ind1][ind2] = weights[ind1][ind2]*100;
    }
    if ( DEBUG ) printf("\n");
    if(ind1 != n-1)
      if(f) fprintf(f, ";\n");
  }
  if(f) fprintf(f, "];\n");
}

void get_gravity_tm(int n, Link*** linkmatrix, int tot_links, FILE* f, 
		    double **demandmatrix){
  int *sum_cap = (int *) malloc( n * sizeof(int)), ind1, ind2;
  bzero(sum_cap, n*sizeof(int));

  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++) {
      Link *l = linkmatrix[ind1][ind2];
      sum_cap[ind1] += (l == NULL)? 0 : l->cap;
    }

  double *weights = (double *) malloc(n * sizeof(double));
  bzero(weights, n*sizeof(double));

  int allsum=0;
  for(ind1=0; ind1<n; ind1++) {
    weights[ind1] = (drand48()*0.9 + .1) *.2;
    allsum += sum_cap[ind1];
  }
  if ( DEBUG ) {
    printf("weights is [");
    for(ind1=0; ind1<n; ind1++)
      printf("%g ", weights[ind1]);
    printf("]\n");
    printf("allsum is %d\n", allsum);
  }

  if(f) fprintf(f, "Demands = [\n");
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++) {
      double demd=0;
      if(ind2 != ind1)
	demd = (weights[ind1] * sum_cap[ind1] * sum_cap[ind2])/ 
	  (allsum - sum_cap[ind1]);
      if(demandmatrix) (demandmatrix)[ind1][ind2] = demd;
      if(f) fprintf(f, " %-9.4f ", demd);
    }
    if(ind1 != n-1)
      if(f) fprintf(f, ";\n");
  }
  if(f) fprintf(f, "];\n");
  free(sum_cap);
  free(weights);
}

/* num_nodes, linkmatrix, num_links */
void print_matlab_topo(int n, Link *** linkmatrix, int tot_links) {
  FILE* matlab_topo = fopen("matlab_topo", "w");

  fprintf (matlab_topo, "n=%d;\nc=%d;\nl=%d;\n", n, n, tot_links);
  fprintf (matlab_topo, "Topo =[");

  int ind1=0, ind2=0;
  for(; ind1< n; ind1++) {
    fprintf(matlab_topo, "\t");
    for(ind2=0; ind2< n; ind2++) {
      Link* l = linkmatrix[ind1][ind2];
      int l_id = (l==NULL)? 0: l->id;
      fprintf(matlab_topo, "%3d ", l_id);
    }
    if (ind1 != n-1) fprintf(matlab_topo, ";\n");
  }

  fprintf(matlab_topo, "];\n");

  fprintf(matlab_topo, "Ends = [\n");
  for(ind1=0; ind1<n; ind1++){
    fprintf(matlab_topo, " %-3d", ind1+1);
    if(ind1 != n-1) 
      fprintf(matlab_topo, ";\n");
  }
  fprintf(matlab_topo, "];\n");

  int* alllink_cap = (int *) malloc(sizeof(int) * tot_links);
  bzero(alllink_cap, tot_links * sizeof(int));
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++){
      Link* l = linkmatrix[ind1][ind2];
      if( l == NULL) continue;
      alllink_cap[l->id] = l->cap;
    }

  fprintf(matlab_topo, "Capy = [\n");
  for(ind1=0; ind1<tot_links; ind1++){
    fprintf(matlab_topo, " %-7d", alllink_cap[ind1+1]);
    if(ind1 != tot_links-1)
      fprintf(matlab_topo, ";\n");
  }
  free(alllink_cap);
  fprintf (matlab_topo, "];\n");

  fclose(matlab_topo);
}



void parse_type_line(char c, char* line, char* type_list, int n){
  char *ct, *ctbig = line;
  while((ct = strstr(ctbig, " ")) != NULL){
    while (isspace(*ct) )  ct++;
    if ( *ct == 0) { 
	printf("reached end of line \n");
        break; 
    }
    int entity = atoi (ct);
    
    if(entity>n || entity<0  || type_list[entity] != ' ') {
	printf("entity: %d, type_list: %c\n",
		entity, type_list[entity]);
      abort();
	}
    
    type_list[entity] = c;

    ctbig = ct;
  }
}


#define RANGESIZE 0.2
void getTMinmargin(int n, double **demands, double margin, double ***out_ptr){
  double **result = *out_ptr;
  int ind1,ind2;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++){
      // proper defn. of margin
      double r1 = (drand48() * (margin-1))+1;

      // strict windows on either sides -- not good
      // double r1 = (drand48() * RANGESIZE)+margin-(RANGESIZE/2);
      int r2 = (drand48() > .5)?1:0;
      double val = demands[ind1][ind2];
      if(DEBUG)printf("val=%g, r1=%g, r2=%d, \n", val, r1, r2);

      if(r2) val /= r1; else val *= r1;
      if(DEBUG) printf("nval=%g\n", val);
      result[ind1][ind2] = val;
    }
}



#define WRITETM 4321
#define WRITEWEIGHT 4322
void __writeweights(int n, double **weights, FILE* fout, int what){
  int ind1, ind2;
  if (what == WRITETM) { fprintf(fout, "Demands = [\n"); }
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++)
      fprintf(fout, "%g ", weights[ind1][ind2]);
    if(what == WRITETM) fprintf(fout, ";");
    fprintf(fout, "\n");
  }
  if (what == WRITETM) { fprintf(fout, "];"); }
}

void writeweights(int n, double **weights, int w, int h, char *prefix, int isFail){
  char cmd[1000];

  if(h==w)    sprintf(cmd, "%s_%d", prefix, w);
  else    sprintf(cmd, "%s_%d_%d", prefix, w, h);

  if ( isFail ) strcat(cmd, ".fail_weights");
  else strcat(cmd, ".weights");

  FILE* fout = fopen(cmd, "w");
  __writeweights(n, weights, fout, WRITEWEIGHT);
  fclose(fout);
}


void printfailflags(int *failflags, int l, FILE* fout){
  int a;
  fprintf(fout, "[");
  for(a=0; a<l; a++)
    if(failflags[a])
      fprintf(fout, "%4d ", a+1);
  fprintf(fout, "] ");
}

void freeallflags(int numlines, int **allfailflags){
  int ind1;
  for(ind1=0; ind1<numlines; ind1++)
    if(allfailflags[ind1]!=NULL) free(allfailflags[ind1]);
  free(allfailflags);
}

void readfailflags_fromline(int **flags_to_be_set, char *line, int fail, int l){
  char temp2[MAXLINELEN];
  char *c_ptr;
  strncpy(temp2, line, MAXLINELEN-1);
  char*tempc = temp2;
  char *temp = (char *)malloc(sizeof(char) * MAXLINELEN);

  // printf("Reading %s\n", temp2);
  int ind2=0;
  while( (c_ptr = (char*)strtok_r(tempc, " ", &temp)) != NULL){
    int d_ = atoi(c_ptr);
    // printf("found %d in line \n", d_);
    if(ind2 > fail) {
      printf("too many links, mismatched fail file\n");
      exit(5);
    }
    if(d_<1 || d_ > l) { printf("bad link %d to fail\n", d_); exit(31); }
    (*flags_to_be_set)[d_-1] = 1;
    ind2++;
    
    if(tempc != NULL) tempc= NULL;
  }
}

int readfailflags(int l, char* dir, int maxlines, int fail, int ***flags_ptr){

  (*flags_ptr) = (int **) malloc( maxlines * sizeof(int*) );
  bzero( *flags_ptr, maxlines * sizeof(int*));
  
  char cmd[1000];
  sprintf(cmd, "tms2cread.pl %s/fail_%d", dir, fail);
  printf("Running %s\n", cmd);
  FILE* fin = popen(cmd, "r");

  char line[MAXLINELEN];
  int ind1=0, linelen, numcols=-1;

  while(fgets(line, MAXLINELEN, fin) != NULL){

    if(ind1<maxlines){
      (*flags_ptr)[ind1] = (int *) malloc(l * sizeof(int));
      bzero( (*flags_ptr)[ind1], l*sizeof(int));
    }else
      break;

    linelen = strlen(line);
    if(linelen == MAXLINELEN) {
      printf("Too long a line: %d\n", linelen); abort();
    }else line[--linelen]=0;

    readfailflags_fromline( &( (*flags_ptr)[ind1] ), line, fail, l);

    if(DEBUG){
      printf("Failflag %d is ", ind1);
      printfailflags( (*flags_ptr)[ind1], l, stdout);
      printf("\n");
    }
    ind1++;
  }

  if(ind1!=maxlines) { printf("\n Couldn't read enough lines. Read %d/%d.\n", ind1, maxlines);}
  fclose(fin);

  return ind1;
}

/* given a bit vector over the links, 
 * modify the topo of links whose bit is set
 * -- serves both bringing down and up!
 */
void modify_topo(Link*** linkmatrix, 
		 char **topomatrix, int*failflags, int n,
		 char replacement, int size_failflags){
  int a, b;
  for(a=0; a<n; a++)
    for(b=0; b<n; b++)
      if(topomatrix[a][b] == '+' ||
	 topomatrix[a][b] == '&' ) /* some sort of link */
	{
	  int l_id = linkmatrix[a][b]->id;
	  if (l_id < 1 || l_id >size_failflags)
	    { printf("bad lid %d\n", l_id); exit(15); }
	  if (failflags[l_id - 1])
	    { if(DEBUG) printf("Taking out %d--%d\n", a, b);
	      topomatrix[a][b] = replacement; }
	}
}

/*
 * reads a standard topo file 
 * num-nodes, 
 * link capacity latency/weight
 * node type list
 * retval == numnodes
 */
int readtopo(FILE* fin, char*** topomatrix_ptr, 
	     Link**** linkmatrix_ptr, char **typelist_ptr, 
	     int* linknum_ptr){
  int linknum_ = 0;
  int numnodes_temp=-1;
  char line[MAXLINELEN];
  int linelen=0;
  int ind1, ind2, a, b;

  char *type_list_, **topomatrix_;
  Link*** linkmatrix_;
  /* read the topo in */
  while(fgets(line, MAXLINELEN, fin) != NULL ) {
    linelen = strlen(line); 
    if(linelen == MAXLINELEN) {
      printf("%s: Too long a line: %d\n", NAME, linelen); abort();
    }else line[--linelen]=0;

    if(line[0] == '#') continue;
    
    if(numnodes_temp == -1 && strstr (line, "NUM-NODES:") == line ){
      numnodes_temp = atoi ( line + strlen("NUM-NODES:") );
      if ( numnodes_temp > MAXNUMNODES) {
	printf("Too many nodes %d. Increase MAXNUMNODES from %d.\n", numnodes_temp, MAXNUMNODES);
	exit(1);
      }
      if(DEBUG) { printf("%s: numnodes got: %d\n", NAME, numnodes_temp); }

      topomatrix_ = (char **) calloc( numnodes_temp, sizeof(char*) );
      linkmatrix_ = (Link ***) calloc( numnodes_temp, sizeof(Link **));

      for(ind1 =0; ind1 < numnodes_temp; ind1++) {
	topomatrix_[ind1] = (char*) calloc(numnodes_temp,  sizeof(char));

	linkmatrix_[ind1] = (Link **) calloc(numnodes_temp, sizeof(Link*));
	bzero(linkmatrix_[ind1], numnodes_temp*sizeof(Link*));

	for(ind2=0; ind2 <numnodes_temp; ind2++)
	  topomatrix_[ind1][ind2] = '-';
	topomatrix_[ind1][ind1] = 'X';
      }


      type_list_ = (char *) calloc(numnodes_temp, sizeof(char));
      for(ind1=0; ind1<numnodes_temp; ind1++)
	type_list_[ind1] = ' ';

      if(DEBUG) print_topo_matrix(topomatrix_, numnodes_temp);
    }else if(numnodes_temp>-1 && strstr(line, "dummy") == line) {
      parse_type_line('d', line, type_list_, numnodes_temp);
    }else if(numnodes_temp>-1 && strstr(line, "ingress") == line) {
      parse_type_line('i', line, type_list_, numnodes_temp);
    }else if(numnodes_temp>-1 && strstr(line, "egress") == line) {
      parse_type_line('e', line, type_list_, numnodes_temp);
    }
    else if (numnodes_temp>-1 && 
	     strstr(line, "LINK:") == line){
      float lat;
      char cap[10];
      char* ctmp = line + strlen("LINK:");

      if ( sscanf(ctmp, " %d %d %s %f", &a, &b, cap, &lat) != 4){
	printf("bad line in topo file: [%s]\n", line);
	exit(343);
      }
      
      if(DEBUG)
	printf("%s: topo entry[%d][%d] set, cap: [%s], lat: %f\n", 
	     NAME, a, b, cap, lat);

      if(! (is_in(a, 0, numnodes_temp-1) && is_in(b, 0, numnodes_temp-1))) {
	printf("%s: bad a: %d, b: %d\n", NAME, a, b); abort();
      }
      
      topomatrix_[a][b] = '+';
      topomatrix_[b][a] = '+';
      Link *l1 = (Link *) malloc(sizeof(Link));
      Link *l2 = (Link *) malloc(sizeof(Link));

      l1->id = ++linknum_;
      l2->id = ++linknum_;
      if( !strcmp(cap, "CC") ){
	l1->cap = l2->cap = CC_CAP;
      }else if( !strcmp(cap, "CP") ){
	l1->cap = l2->cap = CP_CAP;
      }else if( !strcmp(cap, "PP") ){
	l1->cap = l2->cap = PP_CAP;
      }else {
	l1->cap = l2->cap = atof(cap);
      }

      l1->lat = l2->lat = (double)lat;
      linkmatrix_[a][b] = l1;
      linkmatrix_[b][a] = l2;

    }else {
      if(DEBUG)      printf("%s: Ignoring [%s]\n", NAME, line);
    }
  }
  *(topomatrix_ptr) = topomatrix_;
  *(linkmatrix_ptr) = linkmatrix_;
  *(typelist_ptr) = type_list_;
  *(linknum_ptr) = linknum_;

  return numnodes_temp;
}



#define READTM  3000
#define READWT  3001
#define READFL  3002 /* weights optimized for failure */

void __readTM_WTS(int n, double **demandmatrix, FILE* fin)
{
  char line[MAXLINELEN];
  int ind1=0, linelen;
  int numcols=-1;
  

  while(fgets(line, MAXLINELEN, fin) != NULL){
    linelen = strlen(line);
    if(linelen == MAXLINELEN) {
      printf("Too long a line: %d\n", linelen); abort();
    }else line[--linelen]=0;
    
    // printf("Parsing line [%s]\n", line);
    
    char temp[MAXLINELEN];
    char *temp2 = (char *)malloc(sizeof(char) * MAXLINELEN);
    char *c_ptr;

    strcpy(temp, line);
    char*tempc = temp;
    int ind2=0;
    while( (c_ptr = (char*)strtok_r(tempc, " ", &temp2)) != NULL){
      double d_ = atof(c_ptr);
      // printf("%d:%d val is %g\n", ind1, ind2, d_);
      if(ind2 > n || ind1>n) {
	printf("too many fields, mismatched topo-traf?\n");
	exit(5);
      }
      demandmatrix[ind1][ind2++] = d_;
      if(tempc != NULL) tempc= NULL;
    }
    if(numcols == -1) numcols = ind2;
    else if(numcols != ind2) { printf("cols mismatch\n"); exit(8);}
    ind1++;
  }
  if( ind1 != numcols || ind1 != n) 
    { printf("rows mismatch found: %d has to be %d cols: %d\n", ind1, n, numcols); exit(9); }
  
}


// used for both demands and weights
void read2dArray_file(int n, char *filename, double **arr){
  char cmd[1000];
  sprintf(cmd, "tms2cread.pl %s", filename);

  FILE*fin = popen(cmd, "r");
  if( fin == NULL ) { printf("Cannot exec cmd:[%s]\n", cmd); exit(7); }

  __readTM_WTS(n, arr, fin);

  fclose(fin);  
}


void readTMs(int n, char *tms_dir, int which, double **demandmatrix, char *tag){
  char cmd [1000];
  sprintf(cmd, "tms2cread.pl %s/%s_%d.m", tms_dir, tag, which);

  FILE* fin = popen(cmd, "r");
  if( fin == NULL ) { printf("Cannot exec cmd:[%s]\n", cmd); exit(7); }

  __readTM_WTS(n, demandmatrix, fin);

  fclose(fin);  
}


void printutilvalues(int n, int l, Link***linkmatrix, 
		     double **utils, char **topomatrix){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(topomatrix[ind1][ind2] == '+'){
	int l_ = linkmatrix[ind1][ind2]->id;
	printf("%7.4g %3d [%d-->%d] cap %d\n", utils[ind1][ind2], l_, ind1, ind2, linkmatrix[ind1][ind2]->cap);
      }
}


void set_flags(char *c1, int* flags, int n){
  char *temp = (char *) malloc(PREDSTRING_LEN * sizeof(char));
  char *temp2 = temp;
  char *bootemp = (char *) malloc( PREDSTRING_LEN * sizeof(char));
  char *c_ptr, *bootemp2 = bootemp;
  strncpy(bootemp, c1, PREDSTRING_LEN-1);
  while( (c_ptr = (char *)strtok_r(bootemp2, ",", &temp)) != NULL){
    int x_ = atoi(c_ptr);
    if(x_ > n -1 ) { printf ("boo string %s in %s\n", c_ptr, c1); exit(3); }
    flags[x_] = 1;
    if(bootemp2 != NULL) bootemp2=NULL;
  }
  free (temp2);
  free (bootemp);
}


void concat_pred_lists(char *c1, char *c2, char **dest){
  if(DEBUG)  printf("concat [%s] [%s]\n", c1, c2);
  int *flags=(int*)malloc(MAXNUMNODES*sizeof(int));;
  bzero(flags, MAXNUMNODES * sizeof(int));
  
  set_flags(c1, flags, MAXNUMNODES);
  set_flags(c2, flags, MAXNUMNODES);

  char temps[100];
  bzero(temps, 100*sizeof(char));
  int ind1;
  for(ind1=0; ind1<MAXNUMNODES; ind1++)
    if(flags[ind1] == 1){
      if(strlen(temps) > 0)
	sprintf(temps+strlen(temps), ",");
      sprintf(temps+strlen(temps), "%d", ind1);
    }

  if(strlen(temps) > PREDSTRING_LEN)
    {printf ("after concat [%s] is too big\n", temps); exit(4);}
  strcpy(*dest, temps);
  free (flags);
}

void clearfloydmem(int n, char***succlist, double**distmatrix){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++)
      free(succlist[ind1][ind2]);
    free(succlist[ind1]);
    free(distmatrix[ind1]);
  }
  free(distmatrix);
  free(succlist);
}


void printweights_matrix(int n, Link ***linkmatrix) {
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++)
      if(linkmatrix[ind1][ind2] != NULL)
	printf("%5.4g ", (linkmatrix[ind1][ind2]->lat));
      else
	printf("%5s ", "NA");
    printf("\n");
  }
}


double costfn(double util) {
  double retval;
  if (util <= (1.0/3))
    retval= util;
  else if (util <= (2.0/3))
    retval= (1.0/3) + 3 * (util - (1.0/3));
  else if (util <= (9.0/10))
    retval= (4.0/3) + 10 * (util - (2.0/3));
  else if (util <= 1.0)
    retval= (11.0/3) + 70 * (util - (9.0/10));
  else if (util <= (11.0/10))
    retval= (32.0/3) + 500 * (util - 1.0);
  else
    retval= (182.0/3) + 5000 * (util - 1.1);
  return retval;
}

#define FAILURE_BOUNDARY 0.6
double fcostfn(double util){
  double u_ = util/FAILURE_BOUNDARY;
  return costfn(u_);
}


#define MAXWT      20 /* from fortz et. al. */
#define MAXITER    5000 /* from fortz et. al. */
#define RANDWTS    100
#define INVCAPWTS  101
#define UNITWTS    102

void setweights(int n, Link*** linkmatrix, int type, int maxwt){
  int ind1, ind2;

  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(linkmatrix[ind1][ind2] != NULL)
	switch(type){
	case RANDWTS:
	  linkmatrix[ind1][ind2]->lat = (int)(drand48() * (maxwt-1))+1;
	  break;
	case INVCAPWTS:
	  linkmatrix[ind1][ind2]->lat = (int)(CC_CAP/(linkmatrix[ind1][ind2]->cap));
	  break;
	case UNITWTS:
	  linkmatrix[ind1][ind2]->lat = 1;
	  break;
	default:
	  printf("unknow wt type %d\n", type);
	  exit(14);
	}
  /* check */
  if(DEBUG) printweights_matrix(n, linkmatrix);
}




/* somewhat vague, but the weights are hidden inside linkmatrix[a][b]->lat */
int floydwarshall(int n, char **topomatrix, Link*** linkmatrix,
		  int switch_get_equal_wt_paths, double ***dist_ptr,
		  char ****succ_ptr){
  int ind1, ind2, numiter;

  /* shortest distance matrix */
  double** distances, **distances_t_;
  char*** successors, ***successors_t_;

  distances = (double **) calloc(n, sizeof(double));
  distances_t_ = (double **) calloc(n, sizeof(double));

  /* successors strings "a,b,c" => multiple equivalent upstream nodes */
  successors = (char ***) calloc(n, sizeof(char**));
  successors_t_ = (char ***) calloc(n, sizeof(char**));

  for(ind1=0; ind1<n; ind1++) {
    distances[ind1] = (double *) calloc(n, sizeof(double));
    bzero(distances[ind1], n * sizeof(double)); /* 0 => no path! */

    distances_t_[ind1] = (double *) calloc(n, sizeof(double));
    bzero(distances_t_[ind1], n * sizeof(double)); /* 0 => no path! */

    successors[ind1] = (char **) calloc(n, sizeof(char*));
    bzero(successors[ind1], n * sizeof(char*)); /* not set yet */

    successors_t_[ind1] = (char **) calloc(n, sizeof(char*));
    bzero(successors_t_[ind1], n * sizeof(char*)); /* not set yet */

    for(ind2=0; ind2<n; ind2++){
      int x_ = PREDSTRING_LEN * sizeof(char);
      successors_t_[ind1][ind2] = 
	(char *) malloc(x_);
      successors[ind1][ind2] = 
	(char *) malloc(x_);

      bzero(successors_t_[ind1][ind2], x_);
      bzero(successors[ind1][ind2], x_);
    }
  }

  /* init distances and successors */
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(topomatrix[ind1][ind2] == '+'){
	distances[ind1][ind2] = linkmatrix[ind1][ind2]->lat;
	if(DEBUG)
	printf ("init dist %d:%d is %g\n", ind1, ind2, distances[ind1][ind2]);
	char temp[10];
	sprintf(temp, "%d", ind2);
	strncpy(successors[ind1][ind2], temp, PREDSTRING_LEN-1);
      }else {
	distances[ind1][ind2] = -1;
	sprintf(successors[ind1][ind2], "none");
      }

  if(DEBUG) {
    printf("After init\n\n\n");
    print_dist_pred(n, distances, successors);
  }

  for(numiter=0; numiter<n; numiter++){
    for(ind1=0; ind1<n; ind1++)
      for(ind2=0; ind2<n; ind2++){
	double x_ = distances[ind1][numiter] + distances[numiter][ind2];

	if(ind1!=ind2 &&
	   distances[ind1][numiter] >0 &&
	   distances[numiter][ind2] >0 && 
	   (distances[ind1][ind2] == -1 || 
	     distances[ind1][ind2] >= x_ 
	    )){

	  /* new values */
	  if(distances[ind1][ind2] == x_  && switch_get_equal_wt_paths){
	    concat_pred_lists(successors[ind1][ind2], successors[ind1][numiter], 
			      &successors_t_[ind1][ind2]);
	    if(DEBUG)
	      printf("DUP %d->%d, new succ %s\n", ind1, ind2, successors_t_[ind1][ind2]);
	  }
	  else
	    strcpy(successors_t_[ind1][ind2], successors[ind1][numiter]);

	  distances_t_[ind1][ind2] = x_;
	    
	}else {
	  /* copy old values */
	  distances_t_[ind1][ind2] = distances[ind1][ind2];
	  strcpy(successors_t_[ind1][ind2], successors[ind1][ind2]);
	}
      }


    /* swap the matrices */
    double ** dummy = distances;
    distances = distances_t_;
    distances_t_ = dummy;

    char*** c_dummy = successors;
    successors = successors_t_;
    successors_t_ = c_dummy;
  }
  if(DEBUG) {
    printf("At end\n\n\n");
    print_dist_pred(n, distances, successors);
  }
  /* Sanity check - see if graph is connected */
  int retval=0;
  for(ind1=0; ind1<n; ind1++){
    for(ind2=0; ind2<n; ind2++)
      if(ind1 != ind2 && distances[ind1][ind2] < 0) {
	printf("#### graph is disconnected\n");
	retval=1; break;
      }
    if(retval) break;
  }

  *dist_ptr = distances;
  *succ_ptr = successors;

  /* free the temps */
  clearfloydmem(n, successors_t_, distances_t_);
  return retval; /* 0 => success, 1 => disconnection */
}

void unloadweights(double** wts, Link*** links, int n){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(links[ind1][ind2] != NULL)
	wts[ind1][ind2] = links[ind1][ind2]->lat;
}

void loadweights(double** wts, Link*** links, int n){
  int ind1, ind2;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(links[ind1][ind2] != NULL)
	links[ind1][ind2]->lat = wts[ind1][ind2];
}



#define BIGHASHSIZE     16 /* bits */
#define SMALLHASHSIZE   10 /* bits */
void resethash (unsigned int *hashptr, int hashsize ){
  if(hashptr == NULL) { printf ("incoming hash ptr is null\n"); exit(16); }
  bzero(hashptr, (1<<(hashsize-3)));
}
int __checknset (unsigned int *hashptr, int hashsize, int index){
  if (index < 0 || index > (1<<hashsize)-1){
    printf("bad index %d, given hashsize %d\n", index, hashsize); exit(17); 
  }

  unsigned int high = index >> 5;
  unsigned int low = (index - (high << 5));

  if(DEBUG) printf("index %d --> high=%u, low=%u ", index, high, low);

  unsigned int hval = ((hashptr[high]) & ( 1<<low)) >> low;
  hashptr[high] |= (1<<low);

  if(DEBUG) printf("hval=%u\n", hval);
  return hval;
}
void freehash(unsigned int *hashptr){
  free(hashptr);
}
unsigned int* createhash( int hashsize_bits ){
  unsigned int * ptr= (unsigned int *) malloc( (1<<(hashsize_bits-3)) ); /* 2^hashsize_bits bits */
  resethash(ptr, hashsize_bits);
  return ptr;
}
unsigned int fnv32_buf(void *buf, size_t len, unsigned int hval){
    unsigned char *bp = (unsigned char *)buf;	/* start of buffer */
    unsigned char *be = bp + len;		/* beyond end of buffer */

    /*
     * FNV-1 hash each octet in the buffer
     */
    while (bp < be) {
	hval += (hval<<1) + (hval<<4) + (hval<<7) + (hval<<8) + (hval<<24);
	hval ^= (unsigned int)*bp++;
    }

    /* return our new hash value */
    return hval;
}

int dohash(double **wtlist, int n, int l){
  int vecsize=0, ind1, ind2;
  int mywtvector[MAXLINKNUM];
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++)
      if(wtlist[ind1][ind2] > 0)
	mywtvector[vecsize++] = wtlist[ind1][ind2];
  if(vecsize != l){
    printf("mismatch num links has to be %d, is %d\n", l, vecsize); exit(18);
  }
  return fnv32_buf(mywtvector, l*4, 0);
}

int checknset(double **wtlist, int n, int l, unsigned int* hashptr, int hashsize){
  int x_ = dohash(wtlist, n, l) & ((1<<hashsize) - 1);
  //printf("index is %d\n", x_);
  return __checknset(hashptr, hashsize, x_);
}

void setflags_fortree(int n, int target, double*** demands, 
		      char ***successors, char *tree, int howmany){
  int ind1, ind2, tmind;
  bzero(tree, n * sizeof(char));
  for(tmind=0; tmind<howmany; tmind++)
    for(ind1=0; ind1<n; ind1++)
      if( demands[tmind][ind1][target] > 0) 
	tree[ind1] = 1;

  /* check if everybody is already on the tree */
  int everyone_on_tree=1;
  for(ind1=0; ind1<n; ind1++)
    if(ind1 != target && tree[ind1] !=1) { everyone_on_tree=0; break;}
  if(everyone_on_tree) return;


  // drudge thru and add those ferrying demand for others
  int more, *flags = (int*) malloc(n * sizeof(int));
  do{
    more = 0;
    for(ind1=0; ind1<n; ind1++){
      if(tree[ind1] != 1) continue;
      bzero(flags, n*sizeof(int));
      set_flags(successors[ind1][target], flags, n);
      for(ind2=0; ind2<n; ind2++)
	if(flags[ind2] && (tree[ind2] <= 0)) {
	  tree[ind2] = 1;
	  more = 1;  /* check once more because this guy is 
		      * just added onto the tree */
	}
    }
  }while(more);
  free(flags);

  if(DEBUG) printf("Target %d\t Nodes on Tree=", target);
  for(ind1=0; ind1<n; ind1++)
    if(tree[ind1]>0) if(DEBUG) printf("%d, ", ind1);
  if(DEBUG) printf("\n");

  return;
}


double **getutilvalues(double **demandmatrix, char ***succmatrix,
		       int n, Link ***linkmatrix){
  int ind1, ind2, ind3;

  double **utilmatrix = get2ddouble(n);
  double *subload = (double *) malloc(n * sizeof(double));
  int** Tflags = (int **) malloc(n * sizeof(int**));
  int* numflags = (int *) malloc(n*sizeof(int));

  for(ind1=0; ind1<n; ind1++)
    Tflags[ind1] = (int*) malloc(n* sizeof(int));
  
  for(ind1=0; ind1<n; ind1++){

    /* convert all load going towards ind1 into utilization
     * hop-distance on the tree rooted at ind1
     */
    for(ind2=0; ind2<n; ind2++){
      subload[ind2] = demandmatrix[ind2][ind1];
      bzero(Tflags[ind2], n* sizeof(int));
      set_flags(succmatrix[ind2][ind1], Tflags[ind2], n);
      numflags[ind2]=0;
      for(ind3=0; ind3<n; ind3++)
	if(Tflags[ind2][ind3]) numflags[ind2] += 1;
    }
    subload[ind1]=0;

    if(DEBUG) 
    for(ind2=0; ind2<n; ind2++){
      printf("Tflags[%d] \t", ind2);
      for(ind3=0; ind3<n; ind3++) 
	printf(" %3d", Tflags[ind2][ind3]);
      printf("\tTot: %d\n", numflags[ind2]);
    }

    /* go ahead and relax */
    double maxremload;
    int pass_=0;
    do{
      if(DEBUG) {
	printf("For %d Pass %d: ", ind1, pass_++);
	for(ind2=0; ind2<n; ind2++)
	  printf(" %g", subload[ind2]);
	printf("\n");
      }
      for(ind2=0; ind2<n; ind2++){
	if(subload[ind2] == 0 || ind2==ind1) continue;
	for(ind3=0; ind3<n; ind3++)
	  if(Tflags[ind2][ind3]) {
	    double l_ = (subload[ind2]/numflags[ind2]);
	    subload[ind3] += l_;
	    utilmatrix[ind2][ind3] += l_;
	  }
	subload[ind2] = 0;
      }
      if(DEBUG) 
	printf("\tPass %d: load sinked at %d is %g\n", 
	       pass_-1, ind1, subload[ind1]);
      subload[ind1] = 0;
      maxremload = -1;
      for(ind2=0; ind2<n; ind2++)
	if(maxremload < subload[ind2]) maxremload = subload[ind2];
    }while(maxremload > 0);
  }

  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++){
      if(linkmatrix[ind1][ind2] == NULL) continue;
      utilmatrix[ind1][ind2] /= (linkmatrix[ind1][ind2]->cap);
    }

  free(subload);
  free(numflags);
  for(ind1=0; ind1<n; ind1++)
    free(Tflags[ind1]);
  free(Tflags);
  return utilmatrix;
}

double **getutilvalues_over_manyTMS(double ***tms, char ***successors, 
				    int  n, Link ***linkmatrix, int howmany){
  int tmind;
  double **avgutilmatrix = get2ddouble(n);
  for(tmind=0; tmind<howmany; tmind++){
    double **u_ = getutilvalues(tms[tmind], successors, n, linkmatrix);
    add2ddouble(n, avgutilmatrix, u_);
    free2ddouble(n, u_);
  }
  divide2ddouble(n, avgutilmatrix, howmany);
  return avgutilmatrix;
}


void printcritical(int l, char*curr_critical_flags, double* util){
  int ind1;
  printf("[");
  for(ind1=0; ind1<l; ind1++)
    if(curr_critical_flags[ind1]==1)
      printf("\t%3d:%6.4g \n", ind1, util[ind1]);
  printf("]\n");
}


double ospfcost(double **demandmatrix, char ***succmatrix, int n,
		Link ***linkmatrix, double (*fnptr)(double) ){
  int ind1, ind2;
  double **utilmatrix = getutilvalues(demandmatrix, succmatrix, n, linkmatrix);
  double totcost = 0.0;

  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++){
      if(linkmatrix[ind1][ind2] == NULL) continue;
      double load_ = utilmatrix[ind1][ind2];
      if(DEBUG)
	printf("%d-->%d load_ %g, cost: %g\n", 
	       ind1, ind2, load_, fnptr(load_));
      //      totcost += fnptr(load_);  -- srikanth tot-cost fix 09/06/05
      totcost += fnptr(load_) * linkmatrix[ind1][ind2]->cap;
    }
  free2ddouble(n, utilmatrix);

  return totcost;
}

double maxutil_demands(double **demandmatrix, char***succmatrix, 
		       int n, Link*** linkmatrix){
  int ind1, ind2;
  double **utilmatrix = getutilvalues(demandmatrix, succmatrix, n, linkmatrix);
  
  double maxutil=-1.0;
  for(ind1=0; ind1<n; ind1++)
    for(ind2=0; ind2<n; ind2++){
      if(linkmatrix[ind1][ind2] == NULL) continue;
      double load_ = utilmatrix[ind1][ind2];
      if(load_ > maxutil)
	maxutil = load_;
    }
  free2ddouble(n, utilmatrix);
 
  return maxutil;
}

/* Given a current critical set, update it according to the latest weights
 * for the given TM
 * such that the set is no larger than 'size'
 */
void update_critical_set(int n, Link ***linkmatrix, int l,
			 char **topomatrix,
			 double ***tms, int size, 
			 char *curr_critical_flags){
  double *fail_link_maxutil = (double *) malloc(sizeof(double) * l);
  int ind1;

  
  /* compute max-util in topo without this link */
  int *failflags = (int *) malloc( l * sizeof(int));
  double **dist;
  char ***succ;
  for (ind1=0; ind1<l; ind1++){
    bzero(failflags, l * sizeof(int));
    failflags[ind1] = 1;
    modify_topo(linkmatrix, topomatrix, failflags, 
	       n, '&', l);
    
    floydwarshall(n, topomatrix, linkmatrix,
		  1,
		  &dist, &succ);
    fail_link_maxutil[ind1] = maxutil_demands(tms[0], succ, 
					      n, linkmatrix);
    modify_topo(linkmatrix, topomatrix, failflags, 
	       n, '+', l);
    clearfloydmem(n, succ, dist);
  }

  printf("Begin: ");
  printcritical(l, curr_critical_flags, fail_link_maxutil);
  printf("\n");
  free (failflags);

  /* keep adding in new links to the set until no more change   */
  int change;
  do{
    printf("inside\n");
    change = 0;

    double u_inside_min_ = 100;
    int inside_min_id_ = -1;
    double u_bar_ = 0.0;
    int currsize = 0;
    for(ind1=0; ind1<l; ind1++)
      if(curr_critical_flags[ind1] == 1) {
	double u_ = fail_link_maxutil[ind1];
	u_bar_ += u_;
	currsize++;
	if(u_<u_inside_min_) {
	  u_inside_min_ = u_;
	  inside_min_id_ = ind1;
	}
      }

    if(currsize > 0) u_bar_ /= currsize;
    printf("u_bar = %g, currsize= %d\n", u_bar_, currsize);

    double u_outside_max_ = 0.0;
    int outside_max_id_ = -1;
    for(ind1=0; ind1<l; ind1++)
      if(curr_critical_flags[ind1] != 1 &&
	 fail_link_maxutil[ind1] > u_outside_max_){
	u_outside_max_ = fail_link_maxutil[ind1];
	outside_max_id_ = ind1;
      }

    if(currsize < size || 
       u_outside_max_ > u_bar_) {
      /* add it in */
      currsize++;
      printf("Adding in %d\n", outside_max_id_ );
      change = 1;
      curr_critical_flags[outside_max_id_] = 1;

      /* kick out somebody */
      if(currsize > size){
	currsize--;
	if(inside_min_id_< 0) { printf("no min_id\n"); exit(32); }
	else curr_critical_flags[inside_min_id_] = 0;
      }
    }
    
  }while(change == 1);

  printf("End: ");
  printcritical(l, curr_critical_flags, fail_link_maxutil);
  printf("\n");

  free(fail_link_maxutil);
  
}





void computeOSPFCosts( double ***tms, int howmany, int n, 
		       char ***successors, Link***linkmatrix, 
		       double (*fnptr)(double),
		       double *currcost, double *currutil, 
		       char *curr_critical_flags, int l, char **topomatrix,
		       int optimize_over_fails) {
  *currutil = -1;
  *currcost = 0;
  int tmind;  

  /* cost is sum over multiple tms */
  for(tmind=0; tmind<howmany; tmind++){
    (*currcost) += ospfcost(tms[tmind], successors, n, linkmatrix, fnptr);
    double u_temp_ = maxutil_demands(tms[tmind], successors, n, linkmatrix);
    if(u_temp_  > *currutil) *currutil = u_temp_;
  }
  (*currcost) /= howmany;

  if(optimize_over_fails){
    if(DEBUG)printf("%d tms cost is %g\n", howmany, *currcost);
    /* compute failuer cost over the critical set */
    double failure_cost = 0;
    int *failflags = (int*) malloc(l * sizeof(int));
    double **dist;
    char ***succ;
    int ind1, currsize=0;
    for(ind1=0; ind1<l; ind1++)
      if(curr_critical_flags[ind1] != 1) continue;
      else{
	currsize++;
	bzero(failflags, l * sizeof(int));
	failflags[ind1] = 1;
	modify_topo(linkmatrix, topomatrix, failflags,
		    n, '&', l);
	floydwarshall(n, topomatrix, linkmatrix,
		      1,
		      &dist, &succ);
	failure_cost += ospfcost(tms[0], succ, n, linkmatrix, fnptr);
	/* scaled down */
	//failure_cost += ospfcost(tms[0], succ, n, linkmatrix, fcostfn); 

	clearfloydmem(n, succ, dist);
	modify_topo(linkmatrix, topomatrix, failflags,
		    n, '+', l);
      }
    free(failflags);

    if(DEBUG)printf("failure cost on the first tm is %g over %d links\n", 
	   failure_cost, currsize);
    
    if(currsize > 0) {
      failure_cost /= currsize;
      (*currcost) = .5 * ( (*currcost) + failure_cost );
    }
    if(DEBUG)printf("total cost is %g\n", *currcost);
  }
}
		       




/* tms is an array of 'howmany' traffic matrices, over which optimal 
 * weights are to be computed
 */

void setoptimalweights(double ***tms, int n, int l,
		       Link***linkmatrix, char **topomatrix, 
		       int which, int howmany, 
		       int optimize_over_fails, char *outFilePrefix)
{
  struct timeval now;
  gettimeofday(&now, NULL);
  printf("+ BeginOPT %d %d\n", now.tv_sec, now.tv_usec);

  int smallhashsize = 20 * l; /* 20 times the number of arcs */
  int smallhashsizebits = 0;
  while( (1<<smallhashsizebits) < smallhashsize )
    smallhashsizebits++;

  printf("HASH: Big %d bits, Small %d bits\n", BIGHASHSIZE, smallhashsizebits);

  unsigned int* bighash   = createhash(BIGHASHSIZE);
  unsigned int* smallhash = createhash(smallhashsizebits);

  int ind1, ind2, tmind;
  int numiterations = MAXITER;

  /* initialize with rand weights */
  printf("Seeding at %d\n", getpid());
  srand48(getpid());
  setweights(n, linkmatrix, RANDWTS, MAXWT);

  double **currweights = get2ddouble(n);
  double **tempweights = get2ddouble(n);
  double **bestweights = get2ddouble(n);

  char *curr_critical_flags = (char *)malloc(l * sizeof(char));
  bzero(curr_critical_flags, l * sizeof(char));

  double percentsample=.2;
  double currcost, currutil;

  /* set currweights */
  for(ind1=0; ind1<n; ind1++)
    for(ind2=ind1+1; ind2<n; ind2++)
      if(linkmatrix[ind1][ind2] != NULL){
	currweights[ind1][ind2] = linkmatrix[ind1][ind2]->lat;
	currweights[ind2][ind1] = linkmatrix[ind2][ind1]->lat;
      }else{
	currweights[ind1][ind2] = -1;
	currweights[ind2][ind1] = -1;
      }

  /* compute beginning cost */
  double **distances;
  char ***successors;

  /* set state for these weights in the hash table */
  checknset(currweights, n, l, bighash, BIGHASHSIZE);

  /* compute all-pairs shortest paths */
  floydwarshall(n, topomatrix, linkmatrix, 1, &distances, &successors);
  computeOSPFCosts(tms, howmany, n, successors, linkmatrix, costfn,
		   &currcost, &currutil, 
		   curr_critical_flags, l, topomatrix, 
		   optimize_over_fails);
  clearfloydmem(n, successors, distances);
  distances=0;
  successors=0;

  printf("Begin with cost %g, max-util %g\n\n", currcost, currutil);
  int lastiterimproved = -1;


  /* Note: currweights is the local best 
   * tempweights is the new point we are evaluating
   * bestweights is the global best until now
   */

  // initialize bestweights
  copy2ddouble(n, bestweights, currweights);
  double bestcost = currcost, bestutil = currutil;


  // ready to roll!
  for(ind1=0; ind1<numiterations; ind1++) {

    /* new neighborhood exploration -> reset secondary hash table */
    resethash(smallhash, smallhashsizebits);

    int improved=0; /* did local optima improve? */


    /* init temp weights and the linkmatrix */
    copy2ddouble(n, tempweights, currweights);
    loadweights(currweights, linkmatrix, n);

    /* update critical set of link failures */
    if(optimize_over_fails && ind1 > 0 && 
       ind1 % UPDATE_CRITICAL_SET_FREQ == 0)
      update_critical_set(n, linkmatrix, l,
			  topomatrix, tms, CRITICAL_SET_SIZE,
			  curr_critical_flags);

    /* single weight changes */
    int a, b,c;
    for(a=0; a<n; a++)
      for(b=0; b<n; b++){
	if(topomatrix[a][b] != '+') continue;
	for(c=1; c<= MAXWT; c++){
	  if( c == tempweights[a][b] || 
	      drand48() > percentsample) continue;

	  double old = tempweights[a][b];
	  if(DEBUG)
	    printf("changing %d--%d from %g to %d", a, b, old, c);
	  tempweights[a][b] = c;

	  /* have to change the cost elsewhere too, hack */
	  linkmatrix[a][b]->lat = c;
	
	  /* if this wt. vector is not already present in the hash table */
	  if(checknset(tempweights, n, l, bighash, BIGHASHSIZE) == 0 &&
	     checknset(tempweights, n, l, smallhash, smallhashsizebits) == 0) {

	    /* evaluate the cost of the new wt. vector */
	    double ncost, nutil;
	    floydwarshall(n, topomatrix, linkmatrix, 1, &distances, &successors);
	    computeOSPFCosts(tms, howmany, n, successors, linkmatrix, costfn,
			     &ncost, &nutil, curr_critical_flags, l, topomatrix,
			     optimize_over_fails);
	    clearfloydmem(n, successors, distances);
	    distances = 0; successors = 0;

	    if(DEBUG)
	      printf("\t new cost is %g, util is %g", ncost, nutil);
	    
	    if(ncost < currcost ||
	       (ncost == currcost && nutil < currutil)
	       ){
	      if(DEBUG)
		printf("\t Good");
	      copy2ddouble(n, currweights, tempweights);
	      currcost = ncost;
	      currutil = nutil;
	      improved = 1;
	    }
	  }
	  if(DEBUG)
	    printf("\n");
	    
	  /* return tempweights to pristine state */
	  tempweights[a][b] = old;
	  linkmatrix[a][b]->lat = old;
	}
      }
    /* at end, tempweights is the original point, 
     * linkmatrix is same as original, 
     * currweights is the local best from single weight change
     */
    








    /* evenly balancing flows */
    double *totdem = (double*) malloc(n * sizeof(double));
    bzero(totdem, n * sizeof(double));
    for(tmind=0; tmind<howmany; tmind++)
      for(a=0; a<n; a++)
	for(b=0; b<n; b++)
	  totdem[a] += tms[tmind][b][a];

    char *flag_on_tree = (char*) malloc(n * sizeof(char));

    /* get current shortest paths */
    floydwarshall(n, topomatrix, linkmatrix, 1, &distances, &successors);
    if(DEBUG) print_dist_pred(n, distances, successors);

    double **utilmatrix = getutilvalues_over_manyTMS
      (tms, successors, n, linkmatrix, howmany);

    double **storedweights = get2ddouble(n); /* temporary store while probing */
    for(a=0; a<n; a++){
      if(totdem[a] == 0) continue;
       /* for each demand node */

      /* find all nodes ferrying demand TO this node */
      setflags_fortree(n, a, tms, successors, flag_on_tree, howmany);
      for(b=0; b<n; b++){
	if(flag_on_tree[b] <= 0 || b == a) continue;
	if(drand48() > percentsample) continue;
	/* for each node sending traffic to 'a' */
	if(DEBUG)  printf("At tree node %d", b);
	
	double util_thres = (drand48()*.75)+.25;
	if(DEBUG)  printf("\tutilthresh is %g\n", util_thres);

	int *neigh_in_setB = (int *) malloc(n * sizeof(int));
	int *neigh_outside = (int *) malloc(n * sizeof(int));
	bzero(neigh_outside, n * sizeof(int));
	bzero(neigh_in_setB, n * sizeof(int));
	
	/* apply the lightly-loaded constraint first */
	for(c=0; c<n; c++)
	  if(topomatrix[b][c] == '+')
	    if(utilmatrix[b][c] <= util_thres) neigh_in_setB[c]=1;
	    else neigh_outside[c] = 1;


	/* all nodes in set B should have a shorter weighted path to target
	 * than those outside
	 */
	double outside_min_distance = (MAXWT * l)+1;
	for(c=0; c<n; c++){
	  if(neigh_outside[c] != 1) continue;
	  double t_ = linkmatrix[b][c]->lat + distances[c][a];
	  if(t_  < outside_min_distance) outside_min_distance = t_;
	}

	for(c=0; c<n; c++){
	  if(neigh_in_setB[c] != 1) continue;
	  double t_ = linkmatrix[b][c]->lat + distances[c][a];
	  if( t_ > outside_min_distance ) {
	    neigh_outside[c] = 1; neigh_in_setB[c] = 0;
	  }
	}

	double max_inset_distance = -1;
	double min_inset_distance = MAXWT*l + 1;
	for(c=0; c<n; c++){
	  if(neigh_in_setB[c] != 1) continue;
	  double d_ca = distances[c][a];
	  if(c == a) d_ca = 0;
	  if(d_ca < min_inset_distance) min_inset_distance = d_ca;
	}
	if(DEBUG) printf("\t min_inset_distance = %g\n", min_inset_distance);

	/* if set B is not empty */
	if(min_inset_distance < MAXWT*l + 1) {
	  /* throw out neighbors too far away from min */
	  for(c=0; c<n; c++){
	    if(neigh_in_setB[c] != 1) continue;
	    double d_ca = distances[c][a];
	    if(c == a) d_ca = 0;
	    if(d_ca > min_inset_distance + MAXWT-1)
	      { neigh_in_setB[c] = 0; continue; }
	    else if(d_ca > max_inset_distance)
	      max_inset_distance = d_ca;
	  }
	  if(DEBUG) printf("\t max_inset_distance = %g\n", max_inset_distance);

	  /* if set B has more than one element */
	  if( max_inset_distance > -1 && 
	      max_inset_distance != min_inset_distance ) {
	    /* remember tempweights is pristine */
	    copy2ddouble(n, storedweights, tempweights); 

	    double tot_desired = max_inset_distance + 1;
	    for(c=0; c<n; c++) {
	      if(neigh_in_setB[c] != 1) continue;
	      double d_ca = distances[c][a];
	      if(c == a) d_ca = 0;
	      double w_ = tot_desired - d_ca;
	      if(w_ < 1.0 || w_ > MAXWT)
		{ printf("boo nwt sucks %g\n", w_); exit(18); }
	      if(DEBUG) 
		printf("\t BAL %d -> %d from %g to %g (%g)\n",
		       b, c, tempweights[b][c], w_, distances[c][a]);
	      tempweights[b][c]  = w_;
	    }

	    double **dist1;
	    char ***succ1;
	    /* if this is a new wt vector */
	    if(checknset(tempweights, n, l, bighash, BIGHASHSIZE) == 0 && 
	       checknset(tempweights, n, l, smallhash, smallhashsizebits) == 0) {
	      loadweights(tempweights, linkmatrix, n);
	      double ncost, nutil;
	      floydwarshall(n, topomatrix, linkmatrix, l, &dist1, &succ1);
	      computeOSPFCosts(tms, howmany, n, succ1, linkmatrix, costfn,
			       &ncost, &nutil,
			       curr_critical_flags, l, topomatrix,
			       optimize_over_fails);
	      clearfloydmem(n, succ1, dist1);
	      if(DEBUG)	      
	      printf("\t new cost:%g, util: %g", ncost, nutil);
	      if(ncost < currcost ||
		 (ncost == currcost &&  nutil < currutil) ){
		if(DEBUG) 
		  printf("\t\t\t Good\n");
		copy2ddouble(n, currweights, tempweights);
		currcost = ncost;
		currutil = nutil;
		improved = 1;
	      }
	    }
	    if(DEBUG) 
	      printf("\n");

	    /* restore things to their pristine state */
	    copy2ddouble(n, tempweights, storedweights);
	    loadweights(tempweights, linkmatrix, n);
	  }
	}
	
	free(neigh_in_setB);
	free(neigh_outside);
      }
    }
    free (totdem);
    free (flag_on_tree);
    free2ddouble(n, utilmatrix);
    free2ddouble(n, storedweights);
    clearfloydmem(n, successors, distances);

















    /* diversify */
    if(
       /* an obvious check, but not in paper, so why bother */
       // (percentsample == 1 && !improved) ||
	(ind1-lastiterimproved > 300 && !improved) ){
      if(DEBUG) printf("Random perturbation needed!\n");
      printf("Perturb- Cost: %g, Util: %g \t", currcost, currutil);
      for(a=0; a<n; a++)
	for(b=0; b<n; b++){

	  /* perturb 10% of the links */
	  if(topomatrix[a][b]!= '+' || drand48()>.1) continue;

	  /* weight change is uniform in [-2, +2] */
	  int perturb_ = (int)(drand48() * 5)-2;
	  currweights[a][b] += perturb_;
	  
	  /* clamp them */
	  if(currweights[a][b] < 1) currweights[a][b] = 1;
	  if(currweights[a][b] > MAXWT) currweights[a][b] = MAXWT;
	}
      loadweights(currweights, linkmatrix, n);
      floydwarshall(n, topomatrix, linkmatrix, 1, &distances, &successors);
      computeOSPFCosts(tms, howmany, n, successors, linkmatrix, costfn,
		       &currcost, &currutil, curr_critical_flags, l, topomatrix,
		       optimize_over_fails);
      clearfloydmem(n, successors, distances);
      printf("----> Cost: %g, Util: %g\n", currcost, currutil);

      percentsample = .1; /* reset, becomes 20% later on */
    }

    if(improved) { 
      lastiterimproved = ind1;
      /* reset secondary hash table */;
      resethash(smallhash, smallhashsizebits);
    }

    /* update percentsample */
    if(improved)
      percentsample /= 3;
    else
      percentsample *= 2;
    if(percentsample > 1.0) percentsample = 1.0;
    if(percentsample < 0.01) percentsample = 0.01;

    
    printf("Iter %d: new cost=%g, percentsample=%g\n", 
	   ind1, currcost, percentsample);
    
    /* check if improvement in global cost has happened */
    if(currcost < bestcost ||
       (currcost == bestcost && currutil < bestutil) ) {
      printf("Iter %d, GREAT (%g, %g) --> (%g, %g)\n", 
	     ind1, bestcost, bestutil, currcost, currutil);
      copy2ddouble(n, bestweights, currweights);
      loadweights(bestweights, linkmatrix, n);
      if(DEBUG) printweights_matrix(n, linkmatrix);
      bestcost = currcost;
      bestutil = currutil;
    }
  }
  
  /* compute "best" values */
  loadweights(bestweights, linkmatrix, n);
  printweights_matrix(n, linkmatrix);
  
  printf("TMS:%d-%d Ended with cost %g, max-util=%g\n", 
	 which, which+howmany-1, bestcost, bestutil);

  writeweights(n, bestweights, which, which+howmany-1, outFilePrefix, optimize_over_fails);
  free2ddouble(n, currweights);
  free2ddouble(n, tempweights);
  free2ddouble(n, bestweights);
  freehash(bighash);
  freehash(smallhash);

  gettimeofday(&now, NULL);
  printf("+ EndOPT %d %d\n", now.tv_sec, now.tv_usec);

}
